@@ -10,6 +10,7 @@ import openpaperwork_core.deps
 LOGGER = logging.getLogger(__name__)
 
 PDF_FILENAME = 'doc.pdf'
+PASSWD_FILENAME = 'passwd.txt'
 
 PDF_RENDER_FACTOR = 4
 
@@ -138,7 +139,6 @@ class PdfPageMapping(object):
         self.reverse_mapping = {p: p for p in range(0, self.real_nb_pages)}
         now = datetime.datetime.now().timestamp()
         self.page_mtimes = {p: now for p in range(0, self.real_nb_pages)}
-
         if self.core.call_success("fs_exists", self.map_url) is None:
             self.update_nb_pages()
             return
@@ -322,6 +322,8 @@ class PdfPageMapping(object):
 
 
 class Plugin(openpaperwork_core.PluginBase):
+    PRIORITY = 20
+
     def __init__(self):
         super().__init__()
         # we cache the hash of PDF files since they never change
@@ -331,6 +333,7 @@ class Plugin(openpaperwork_core.PluginBase):
         self.cache_nb_pages = {}
 
         self.cache_mappings = {}
+        self.cache_passwords = {}
 
     def get_interfaces(self):
         return [
@@ -364,7 +367,7 @@ class Plugin(openpaperwork_core.PluginBase):
             {
                 'interface': 'pillow',
                 'defaults': [
-                    'paperwork_backend.pillow.img',
+                    'openpaperwork_core.pillow.img',
                     'paperwork_backend.pillow.pdf',
                 ],
             },
@@ -372,6 +375,10 @@ class Plugin(openpaperwork_core.PluginBase):
                 'interface': 'poppler',
                 'defaults': ['paperwork_backend.poppler.memory'],
             },
+            {
+                'interface': 'urls',
+                'defaults': ['openpaperwork_core.urls'],
+            },
         ]
 
     def _get_pdf_url(self, doc_url):
@@ -382,6 +389,20 @@ class Plugin(openpaperwork_core.PluginBase):
             return None
         return pdf_url
 
+    def _get_pdf_password(self, doc_url):
+        password = self.cache_passwords.get(doc_url, None)
+        if password is not None:
+            return password
+
+        passwd_url = self.core.call_success(
+            "fs_join", doc_url, PASSWD_FILENAME
+        )
+        if self.core.call_success("fs_exists", passwd_url):
+            with self.core.call_success("fs_open", passwd_url, "r") as fd:
+                password = fd.read().strip()
+        self.cache_passwords[doc_url] = password
+        return password
+
     def _get_page_mapping(self, doc_url):
         if doc_url in self.cache_mappings:
             return self.cache_mappings[doc_url]
@@ -393,8 +414,12 @@ class Plugin(openpaperwork_core.PluginBase):
         pdf_url = self._get_pdf_url(doc_url)
         if pdf_url is None:
             return (None, None)
-        LOGGER.info("Opening %s", pdf_url)
-        doc = self.core.call_success("poppler_open", pdf_url)
+
+        password = self._get_pdf_password(doc_url)
+        LOGGER.info("Opening %s (password=%s)", pdf_url, bool(password))
+        doc = self.core.call_success(
+            "poppler_open", pdf_url, password=password
+        )
         return (pdf_url, doc)
 
     def is_doc(self, doc_url):
@@ -403,24 +428,27 @@ class Plugin(openpaperwork_core.PluginBase):
             return None
         return True
 
-    def doc_get_pdf_url_by_url(self, doc_url):
+    def doc_get_pdf_url_by_url(self, doc_url, write=False):
+        if write:
+            return self.core.call_success("fs_join", doc_url, PDF_FILENAME)
         return self._get_pdf_url(doc_url)
 
-    def doc_internal_get_hash_by_url(self, out: list, doc_url):
+    def flush_doc_cache(self, doc_url):
+        self.cache_hash.pop(doc_url, None)
+        self.cache_passwords.pop(doc_url, None)
+        self.cache_mappings.pop(doc_url, None)
+        self.cache_nb_pages.pop(doc_url, None)
+
+    def doc_get_hash_by_url(self, doc_url):
         pdf_url = self._get_pdf_url(doc_url)
         if pdf_url is None:
             return
-        mapping = self._get_page_mapping(doc_url)
-        h = mapping.get_map_hash()
-        if h is not None:
-            out.append(h)
 
         # cache the hash of doc.pdf to speed up imports
-        cache_key = "hash_{}".format(doc_url)
-        if cache_key not in self.cache_hash:
+        if doc_url not in self.cache_hash:
             h = self.core.call_success("fs_hash", pdf_url)
-            self.cache_hash[cache_key] = h
-        out.append(self.cache_hash[cache_key])
+            self.cache_hash[doc_url] = h
+        return self.cache_hash[doc_url]
 
     def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
         pdf_url = self._get_pdf_url(doc_url)
@@ -434,11 +462,10 @@ class Plugin(openpaperwork_core.PluginBase):
         out.append(page_hash)
 
         # cache the hash of doc.pdf to speed up imports
-        cache_key = "hash_{}".format(doc_url)
-        if cache_key not in self.cache_hash:
+        if doc_url not in self.cache_hash:
             h = self.core.call_success("fs_hash", pdf_url)
-            self.cache_hash[cache_key] = h
-        out.append(self.cache_hash[cache_key])
+            self.cache_hash[doc_url] = h
+        out.append(self.cache_hash[doc_url])
 
     def doc_internal_get_mtime_by_url(self, out: list, doc_url):
         pdf_url = self._get_pdf_url(doc_url)
@@ -527,8 +554,18 @@ class Plugin(openpaperwork_core.PluginBase):
                 return None
             else:
                 original_page_idx = page_idx
+
+        password = self._get_pdf_password(doc_url)
+        if password is not None:
+            password = password.encode("utf-8").hex()
+
         # same URL used in browsers
-        return "{}#page={}".format(pdf_url, str(original_page_idx + 1))
+        url = self.core.call_success(
+            "url_args_join", pdf_url,
+            page=str(original_page_idx + 1),
+            password=password
+        )
+        return url
 
     @staticmethod
     def _custom_split(input_str, input_rects, splitter, log_txt):
@@ -692,7 +729,7 @@ class Plugin(openpaperwork_core.PluginBase):
             "mainloop_execute", self._page_get_boxes_by_url, doc_url, page_idx
         )
 
-    def doc_pdf_import(self, src_file_uri):
+    def doc_pdf_import(self, src_file_uri, password=None):
         (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
 
         # just to be safe
@@ -703,9 +740,18 @@ class Plugin(openpaperwork_core.PluginBase):
         self.core.call_success("fs_mkdir_p", doc_url)
         self.core.call_success("fs_copy", src_file_uri, pdf_url)
 
+        if password is not None:
+            passwd_url = self.core.call_success(
+                "fs_join", doc_url, PASSWD_FILENAME
+            )
+            with self.core.call_success("fs_open", passwd_url, "w") as fd:
+                fd.write(password)
+
         try:
             # check the PDF is readable
-            doc = self.core.call_success("poppler_open", pdf_url)
+            doc = self.core.call_success(
+                "poppler_open", pdf_url, password=password
+            )
             exc_info = None
         except Exception as exc:
             doc = None
@@ -717,6 +763,7 @@ class Plugin(openpaperwork_core.PluginBase):
             else:
                 LOGGER.error("Failed to read %s", pdf_url, exc_info=exc_info)
             self.core.call_success("fs_rm_rf", doc_url, trash=False)
+            return (None, None)
 
         return (doc_id, doc_url)
 
@@ -797,12 +844,11 @@ class Plugin(openpaperwork_core.PluginBase):
             original_source_page_idx = source_mapping.get_original_page_idx(
                 target_source_page_idx
             )
-            if original_source_page_idx is None:
-                # this page is not handled by us ; still, we must shift
-                # all our pages
-                original_source_page_idx = target_source_page_idx
+            # if original_source_page_idx is None means
+            # this page is not handled by us ; still, we must shift
+            # all our pages
             LOGGER.info(
-                "- Removing page (original=%d, target=%d) from %s",
+                "- Removing page (original=%s, target=%d) from %s",
                 original_source_page_idx, target_source_page_idx,
                 source_doc_url
             )
@@ -816,39 +862,45 @@ class Plugin(openpaperwork_core.PluginBase):
             dest_mapping.make_room_for_target_page(target_dest_page_idx)
 
         if source_doc_url == dest_doc_url:
-            assert(source_mapping == dest_mapping)
-            LOGGER.info(
-                "New mapping: %s: original=p%d --> target=p%d",
-                source_doc_url, original_source_page_idx, target_dest_page_idx
-            )
-            source_mapping.set_mapping(
-                original_source_page_idx, target_dest_page_idx
-            )
-            source_mapping.update_nb_pages()
+            assert(source_mapping is dest_mapping)
+            if original_source_page_idx is not None:
+                LOGGER.info(
+                    "New mapping: %s: original=p%d --> target=p%d",
+                    source_doc_url, original_source_page_idx,
+                    target_dest_page_idx
+                )
+                source_mapping.set_mapping(
+                    original_source_page_idx, target_dest_page_idx
+                )
+
+                source_mapping.update_nb_pages()
         elif source_is_pdf:
-            # export the PDF page as an image file
-            # it relies on other model plugins (interface 'page_img'), but they
-            # can't be declared as dependencies, as we do provide 'page_img'
-            # too. It would make a dependency loop.
+            if original_source_page_idx is not None:
+                # export the PDF page as an image file
+                # it relies on other model plugins (interface 'page_img'), but
+                # they can't be declared as dependencies, as we do provide
+                # 'page_img' too. It would make a dependency loop.
 
-            # we are a low priority plugin: other plugins should
-            # already have made room for our page if required
+                # we are a low priority plugin: other plugins should
+                # already have made room for our page if required
 
-            source_img = "{}#page={}".format(
-                source_pdf_url, str(original_source_page_idx + 1)
-            )
-            LOGGER.info("Generating image from %s", source_img)
-            source_img = self.core.call_success("url_to_pillow", source_img)
+                source_img = "{}#page={}".format(
+                    source_pdf_url, str(original_source_page_idx + 1)
+                )
+                LOGGER.info("Generating image from %s", source_img)
+                source_img = self.core.call_success(
+                    "url_to_pillow", source_img
+                )
 
-            # since PDF are not writable by themselves, we can call
-            # page_get_img_url(write=True). We are sure that our
-            # implementation of this method won't reply
-            dest_img = self.core.call_success(
-                "page_get_img_url", dest_doc_url, target_dest_page_idx,
-                write=True
-            )
-            LOGGER.info("Writting page image back as %s", dest_img)
-            self.core.call_success("pillow_to_url", source_img, dest_img)
+                # since PDF are not writable by themselves, we can call
+                # page_get_img_url(write=True). We are sure that our
+                # implementation of this method won't reply
+                dest_img = self.core.call_success(
+                    "page_get_img_url", dest_doc_url, target_dest_page_idx,
+                    write=True
+                )
+                LOGGER.info("Writting page image back as %s", dest_img)
+                self.core.call_success("pillow_to_url", source_img, dest_img)
 
         if source_is_pdf:
             source_mapping.save()
